#include <cgaborator.h>
#include <samplerate.h>
#include <memory>
#include <unordered_map>
#include <vector>
#include "lemire.h"
#include "hibiki.h"


void audio_multiple_channels_to_mono(float* buffer, size_t buffer_len, int channels) {
    float f;
    switch(channels) {
        case 1:
            break;
        case 2:
            for (int i = 0; i < buffer_len; i += 2){
                buffer[i/2] = buffer[i] + buffer[i+1];
            }
            break;
        default:
            for (int i = 0; i < buffer_len; i += channels){
                f = buffer[i];
                for (int j = 1; j < channels; ++j) {
                    f += buffer[i + j];
                }
                buffer[i/channels] = f / (float)channels;
            }
    }
}

void write_uint32_le(std::vector<uint8_t>& d, uint32_t v) {
    d.push_back(v & 0xFF);
    d.push_back((v >> 8ul) & 0xFF);
    d.push_back((v >> 16ul) & 0xFF);
    d.push_back((v >> 24ul) & 0xFF);
}
void write_uint32_be(std::vector<uint8_t>& d, uint32_t v) {
    d.push_back((v >> 24ul) & 0xFF);
    d.push_back((v >> 16ul) & 0xFF);
    d.push_back((v >> 8ul) & 0xFF);
    d.push_back(v & 0xFF);
}
void write_uint64_le(std::vector<uint8_t>& d, uint64_t v) {
    d.push_back(v & 0xFF);
    d.push_back((v >> 8ul) & 0xFF);
    d.push_back((v >> 16ul) & 0xFF);
    d.push_back((v >> 24ul) & 0xFF);
    d.push_back((v >> 32ul) & 0xFF);
    d.push_back((v >> 40ul) & 0xFF);
    d.push_back((v >> 48ul) & 0xFF);
    d.push_back((v >> 56ul) & 0xFF);
}
void write_uint64_be(std::vector<uint8_t>& d, uint64_t v) {
    d.push_back((v >> 56ul) & 0xFF);
    d.push_back((v >> 48ul) & 0xFF);
    d.push_back((v >> 40ul) & 0xFF);
    d.push_back((v >> 32ul) & 0xFF);
    d.push_back((v >> 24ul) & 0xFF);
    d.push_back((v >> 16ul) & 0xFF);
    d.push_back((v >> 8ul) & 0xFF);
    d.push_back(v & 0xFF);
}

class EventPointProcessor {
private:
    const int bandsPerOctave = PANAKO_TRANSF_BANDS_PER_OCTAVE;
    const int audioBlockSize = PANAKO_AUDIO_BLOCK_SIZE;
    const int maxFilterWindowSizeFrequency = PANAKO_FREQ_MAX_FILTER_SIZE;
    const int maxFilterWindowSizeTime = PANAKO_TIME_MAX_FILTER_SIZE;
public:
    EventPointProcessor(int sampleRate, int channels, int timeResolutionInMs) : sampleRate(sampleRate), channels(channels), timeResolution(double(PANAKO_SAMPLE_RATE) * (double(timeResolutionInMs) / 1000)){
        gaborator = gaborator_initialize(
                audioBlockSize,
                PANAKO_SAMPLE_RATE,
                bandsPerOctave,
                PANAKO_TRANSF_MIN_FREQ,
                PANAKO_TRANSF_REF_FREQ,
                PANAKO_TRANSF_MAX_FREQ,
                timeResolution
        );

        printf("hibiki: convert from %d Hz to %d Hz, %d to 1 channel(s)\n", sampleRate, PANAKO_SAMPLE_RATE, channels);
        src = src_new(SRC_LINEAR, 1, nullptr);
        bandNumber = gaborator_number_of_bands(gaborator);
        latency = gaborator_analysis_support(gaborator);
        maxFilterVertical = lemire_max_clamped_init(maxFilterWindowSizeFrequency, bandNumber);
        if (maxFilterVertical == nullptr){
            abort();
        }

        magnitudes.resize(maxFilterWindowSizeTime);

        maxMagnitudes.resize(maxFilterWindowSizeTime);
        for (auto& v : maxMagnitudes) {
            v.resize(bandNumber);
        }
        maxHorizontal.resize(bandNumber);
    }

    void Process(float* data, size_t data_size) {
        if (data_size % channels != 0){
            abort();
        }
        sampleCount += data_size;
        //printf("processing %p %d elements\n", data, data_size);
        audio_multiple_channels_to_mono(data, data_size, channels);
        FeedSampler(data, data_size / channels);
    }

    void FeedSampler(float* data, size_t data_size){
        //printf("sampler: processing %p %d elements\n", data, data_size);
        bool eof = (data_size == 0);
        if(sampleRate == PANAKO_SAMPLE_RATE){
           if(!eof){
                audioSamplesOut.insert(audioSamplesOut.end(), data, data + data_size);
           }
        }else{
            std::vector<float> buf(audioBlockSize * (double(PANAKO_SAMPLE_RATE) / double(sampleRate)) * 1.5);

            //printf("sample: enter %d\n", data_size);
            SRC_DATA entry;

            entry.data_in = data;
            entry.input_frames = data_size;
            entry.end_of_input = eof;

            entry.data_out = buf.data();
            entry.output_frames = buf.size();

            entry.src_ratio = double(PANAKO_SAMPLE_RATE) / double(sampleRate);

            auto err = src_process(src, &entry);
            if(err != 0){
                abort();
            }
            //printf("sample: process used %d gen %d ~buf.size %d\n", entry.input_frames_used, entry.output_frames_gen, buf.size());

            if(entry.output_frames_gen > 0){
                audioSamplesOut.insert(audioSamplesOut.end(), buf.data(), buf.data() + entry.output_frames_gen);
            }
        }

        while ((audioSamplesOut.size() - audioSamplesOutIndex) >= audioBlockSize){
            //printf("sample: transform from %d for size %d\n", audioSamplesOutIndex, audioBlockSize);
            Transform(audioSamplesOut.data() + audioSamplesOutIndex, audioBlockSize);
            audioSamplesOutIndex += audioBlockSize;
        }

        if(eof && (audioSamplesOut.size() - audioSamplesOutIndex) > 0){
            //printf("sample: END transform from %d for size %d\n", audioSamplesOutIndex, audioSamplesOut.size() - audioSamplesOutIndex);

            Transform(audioSamplesOut.data() + audioSamplesOutIndex, audioSamplesOut.size() - audioSamplesOutIndex);
            audioSamplesOut.resize(0);
            audioSamplesOutIndex = 0;
        }
    }

    double blocksToSeconds(uint32_t t) const {
        return double(t)*(double(timeResolution)/double(PANAKO_SAMPLE_RATE)) + double(latency)/double(PANAKO_SAMPLE_RATE);
    }

    void Transform(const float* data, size_t data_size) {

        //printf("transform %p for size %d\n", data, data_size);
        size_t return_size, slice_size;
        float* b_data = gaborator_transform(gaborator, const_cast<float*>(data), data_size, &return_size, &slice_size);

        if(return_size > 0){
            //printf("%d got %p slices of %d size, total %d, count %d\n", callback_data, b_data, slice_size, size, size / slice_size);
            for (auto i = 0; i < return_size; i += slice_size){
                processMagnitudes(b_data + i, slice_size);
            }
        }
    }

    const std::vector<uint8_t>& Finish(bool useCompactHash, bool usePackedPrint) {
        //Flush sampler
        std::vector<float> b(1);
        FeedSampler(b.data(), 0);

        //Flush gaborator
        Transform(nullptr, 0);

        byte_result.clear();


        printf("hibiki: total samples %d\n", sampleCount);
        printf("hibiki: %d event points\n", points.size());

        auto prints = packEventPoints();
        printf("hibiki: %d fingerprints\n", prints.size());
        if (!prints.empty()){
            printf("hibiki: estimated duration %f\n", blocksToSeconds(prints[prints.size() - 1].t1()));
        }
        byte_result.reserve(prints.size() * ((useCompactHash ? 4 : 8) + (usePackedPrint ? 4 : 8)));
        for (auto& print : prints){
            if(useCompactHash){
                write_uint32_be(byte_result, print.compactHash());
            } else {
                write_uint64_be(byte_result, print.hash());
            }

            if(usePackedPrint){
                write_uint32_le(byte_result, (print.t1() & 0b11111111111111111111) | ((print.f1() & 0b111111111111) << 20));
            }else{
                write_uint32_le(byte_result, print.t1());
                write_uint32_le(byte_result, print.f1());
            }
        }

        return byte_result;
    }

    int getBandNumber() const {
        return bandNumber;
    }

    ~EventPointProcessor() {
        gaborator_release(gaborator);
        lemire_max_clamped_free(maxFilterVertical);
        src_delete(src);
    }

private:

    void processMagnitudes(float* data, size_t size) {
        magnitudes[magnitudesIndex] = std::vector<float>(data, data + size);
        //printf("process magnitude %d %d %d %d\n", magnitudesIndex, analysisFrameIndex, size, magnitudes[magnitudesIndex].size());

        //store the frame magnitudes
        previousMagnitudes[analysisFrameIndex] = &magnitudes.at(magnitudesIndex);

        //run a max filter over frequency bins
        lemire_max_clamped_filter(maxFilterVertical, magnitudes[magnitudesIndex].data(), magnitudes[magnitudesIndex].size(), maxMagnitudes[magnitudesIndex].data());

        //store the max filtered frequency bins
        previousMaxMagnitudes[analysisFrameIndex] = &maxMagnitudes.at(magnitudesIndex);

        //find the horizontal maxima
        if(previousMaxMagnitudes.size() == maxFilterWindowSizeTime){
            auto time = analysisFrameIndex - maxFilterWindowSizeTime/2;
            auto& maxFrame = *previousMaxMagnitudes[time];
            auto& frameMagnitudes = *previousMagnitudes[time];

            for (auto frequency = 2; frequency < frameMagnitudes.size()-1; frequency++) {
                auto maxVal = maxFrame[frequency];
                auto currentVal = frameMagnitudes[frequency];

                if (maxVal == currentVal && currentVal != 0) {
                    horizontalFilter(frequency);
                    maxVal = maxHorizontal[frequency];
                    if (currentVal == maxVal) {
                        auto& prevFrameMagnitudes = *previousMagnitudes[time-1];
                        auto& nextFrameMagnitudes = *previousMagnitudes[time+1];

                        //add the magnitude of surrounding bins for magnitude estimates more robust against discretization effects
                        float totalMagnitude = frameMagnitudes[frequency] + prevFrameMagnitudes[frequency] + nextFrameMagnitudes[frequency] +
                        frameMagnitudes[frequency+1] + prevFrameMagnitudes[frequency+1] + nextFrameMagnitudes[frequency+1] +
                        frameMagnitudes[frequency-1] + prevFrameMagnitudes[frequency-1] + nextFrameMagnitudes[frequency-1];

                        points.emplace_back(uint32_t(time), uint32_t(frequency), totalMagnitude);
                    }
                }
            }

            //Remove analysis frames that are not needed any more:
            auto ix = analysisFrameIndex - maxFilterWindowSizeTime + 1;
            previousMaxMagnitudes.erase(ix);
            previousMagnitudes.erase(ix);
        }

        magnitudesIndex++;
        if (magnitudesIndex == maxFilterWindowSizeTime){
            magnitudesIndex = 0;
        }
        analysisFrameIndex++;
    }

    void horizontalFilter(int j) {
        maxHorizontal[j] = -1000;

        auto centerFrameIndex = analysisFrameIndex - maxFilterWindowSizeTime/2;
        auto startFrameIndex = centerFrameIndex - maxFilterWindowSizeTime/2;
        auto stopFrameIndex = centerFrameIndex + maxFilterWindowSizeTime/2;

        // Run a horizontal max filter
        for (auto i = startFrameIndex; i < stopFrameIndex; i++) {
            auto& maxFrame = *previousMaxMagnitudes[i];
            if (maxFrame[j] > maxHorizontal[j]) {
                maxHorizontal[j] = maxFrame[j];
            }
        }
    }

    class EventPoint {
    public:
        EventPoint(uint32_t time, uint32_t frequency, float magnitude) : time(time), frequency(frequency), magnitude(magnitude){

        }

        const uint32_t time; // The time expressed using an analysis frame index.
        const uint32_t frequency; // The frequency expressed using the bin number in the constant Q transform.
        const float magnitude; // The energy value of the element.

        int64_t DeltaTime(const EventPoint& p) const {
            return static_cast<int64_t>(time) - static_cast<int64_t>(p.time);
        }
    };

    class Fingerprint {
    private:
        const EventPoint& p1;
        const EventPoint& p2;
        const EventPoint& p3;
    public:
        Fingerprint(const EventPoint& p1, const EventPoint& p2, const EventPoint& p3) : p1(p1), p2(p2), p3(p3) {
            if(p2.DeltaTime(p1) < 0){
                abort();
            }
            if(p3.DeltaTime(p2) < 0){
                abort();
            }
        }


        uint32_t t1() const {
            return p1.time;
        }
        uint32_t t2() const {
            return p2.time;
        }
        uint32_t t3() const {
            return p3.time;
        }

        uint32_t f1() const {
            return p1.frequency;
        }
        uint32_t f2() const {
            return p2.frequency;
        }
        uint32_t f3() const {
            return p3.frequency;
        }

        float m1() const {
            return p1.magnitude;
        }
        float m2() const {
            return p2.magnitude;
        }
        float m3() const {
            return p3.magnitude;
        }
        
        uint64_t hash() const {
            uint64_t f1LargerThanF2 = f1() > f2();
            uint64_t f2LargerThanF3 = f2() > f3();
            uint64_t f3LargerThanF1 = f3() > f1();


            uint64_t m1LargerThanM2 = m1() > m2();
            uint64_t m2LargerThanM3 = m2() > m3();
            uint64_t m3LargerThanM1 = m3() > m1();

            int64_t diffT1T2 = int64_t(t2()) - int64_t(t1());
            int64_t diffT3T2 = int64_t(t3()) - int64_t(t2());
                    
            uint64_t dt1t2LargerThant3t2 = diffT1T2 > diffT3T2;

            int64_t diffF2F1 = std::abs(int64_t(f2()) - int64_t(f1()));
            int64_t diffF3F2 = std::abs(int64_t(f3()) - int64_t(f2()));


            uint64_t df1f2LargerThanf3f2 = diffF2F1 > diffF3F2;

            //9 bits f in range( 0 - 512) to 8 bits
            uint64_t f1Range = f1() >> 1;

            //7 bits (0-128) -> 5 bits
            auto df2f1 = diffF2F1 >> 2;
            auto df3f2 = diffF3F2 >> 2;

            //6 bits max
            auto ratioT = uint64_t(double(diffT1T2) / double(diffT3T2) * 64);

            //combine the hash components into a single 64-bit unsigned integer

            //for debugging purposes, these should be inlined later on
            auto clampBits = [](uint64_t v, int n) {
                    return v & ((1 << n) - 1);
            };
            auto shiftBits = [](uint64_t v, int n) {
                return v << n;
            };

            //34 bit output
            // +----+----------------+---|
            // |  # |     field      | n |
            // +----+----------------+---|
            // |    |                |   |
            // |    |                |   |
            // |  0 |     ratioT     | 6 |
            // |    |                |   |
            // |    |                |   |
            // |    |                |   |
            // +----+----------------+---|
            // |  6 |   F1 > f2      | 1 |
            // +----+----------------+---|
            // |  7 |   f2 > f3      | 1 |
            // +----+----------------+---|
            // |  8 |   f3 > F1      | 1 |
            // +----+----------------+---|
            // |  9 |   m1 > m2      | 1 |
            // +----+----------------+---|
            // | 10 |   m2 > m3      | 1 |
            // +----+----------------+---|
            // | 11 |   m3 > m1      | 1 |
            // +----+----------------+---|
            // | 12 |d(T1,t2)>d(t3,t2)| 1 |
            // +----+----------------+----|
            // | 13 |d(F1,f2)>d(f3,f2)| 1 |
            // +----+----------------+---|
            // |    |                |   |
            // |    |                |   |
            // |    |                |   |
            // | 14 |     f1range    | 8 |
            // |    |                |   |
            // |    |                |   |
            // |    |                |   |
            // |    |                |   |
            // +----+----------------+---|
            // |    |                |   |
            // |    |                |   |
            // | 22 |     d(f2,F1)   | 6 |
            // |    |                |   |
            // |    |                |   |
            // |    |                |   |
            // +----+----------------+---|
            // |    |                |   |
            // |    |                |   |
            // | 28 |     d(f3,f2)   | 6 |
            // |    |                |   |
            // |    |                |   |
            // |    |                |   |
            // +----+----------------+---|
            return shiftBits(clampBits(ratioT, 6), 0) |
                     shiftBits(clampBits(f1LargerThanF2, 1), 6) |
                     shiftBits(clampBits(f2LargerThanF3, 1), 7) |
                     shiftBits(clampBits(f3LargerThanF1, 1), 8) |
                     shiftBits(clampBits(m1LargerThanM2, 1), 9) |
                     shiftBits(clampBits(m2LargerThanM3, 1), 10) |
                     shiftBits(clampBits(m3LargerThanM1, 1), 11) |
                     shiftBits(clampBits(dt1t2LargerThant3t2, 1), 12) |
                     shiftBits(clampBits(df1f2LargerThanf3f2, 1), 13) |
                     shiftBits(clampBits(f1Range, 8), 14) |
                     shiftBits(clampBits(uint64_t(df2f1), 6), 22) |
                     shiftBits(clampBits(uint64_t(df3f2), 6), 28);
        }



        uint32_t compactHash() const {
            uint32_t f1LargerThanF2 = f1() > f2();
            uint32_t f2LargerThanF3 = f2() > f3();
            uint32_t f3LargerThanF1 = f3() > f1();


            uint32_t m1LargerThanM2 = m1() > m2();
            uint32_t m2LargerThanM3 = m2() > m3();
            uint32_t m3LargerThanM1 = m3() > m1();

            int64_t diffT1T2 = int64_t(t2()) - int64_t(t1());
            int64_t diffT3T2 = int64_t(t3()) - int64_t(t2());

            uint32_t dt1t2LargerThant3t2 = diffT1T2 > diffT3T2;

            int64_t diffF2F1 = std::abs(int64_t(f2()) - int64_t(f1()));
            int64_t diffF3F2 = std::abs(int64_t(f3()) - int64_t(f2()));


            uint32_t df1f2LargerThanf3f2 = diffF2F1 > diffF3F2;

            //9 bits f in range( 0 - 512) to 6 bits
            uint32_t f1Range = f1() >> 3;

            //7 bits (0-128) -> 5 bits
            auto df2f1 = diffF2F1 >> 2;
            auto df3f2 = diffF3F2 >> 2;

            //6 bits max
            auto ratioT = uint32_t(double(diffT1T2) / double(diffT3T2) * 64);

            //combine the hash components into a single 32-bit unsigned integer

            //for debugging purposes, these should be inlined later on
            auto clampBits = [](uint32_t v, int n) {
                return v & ((1 << n) - 1);
            };
            auto shiftBits = [](uint32_t v, int n) {
                return v << n;
            };

            //32 bit output
            // +----+----------------+---|
            // |  # |     field      | n |
            // +----+----------------+---|
            // |    |                |   |
            // |    |                |   |
            // |  0 |     ratioT     | 6 |
            // |    |                |   |
            // |    |                |   |
            // |    |                |   |
            // +----+----------------+---|
            // |  6 |   F1 > f2      | 1 |
            // +----+----------------+---|
            // |  7 |   f2 > f3      | 1 |
            // +----+----------------+---|
            // |  8 |   f3 > F1      | 1 |
            // +----+----------------+---|
            // |  9 |   m1 > m2      | 1 |
            // +----+----------------+---|
            // | 10 |   m2 > m3      | 1 |
            // +----+----------------+---|
            // | 11 |   m3 > m1      | 1 |
            // +----+----------------+---|
            // | 12 |d(T1,t2)>d(t3,t2)| 1 |
            // +----+----------------+----|
            // | 13 |d(F1,f2)>d(f3,f2)| 1 |
            // +----+----------------+---|
            // |    |                |   |
            // |    |                |   |
            // | 14 |     f1range    | 6 |
            // |    |                |   |
            // |    |                |   |
            // |    |                |   |
            // +----+----------------+---|
            // |    |                |   |
            // |    |                |   |
            // | 20 |     d(f2,F1)   | 6 |
            // |    |                |   |
            // |    |                |   |
            // |    |                |   |
            // +----+----------------+---|
            // |    |                |   |
            // |    |                |   |
            // | 26 |     d(f3,f2)   | 6 |
            // |    |                |   |
            // |    |                |   |
            // |    |                |   |
            // +----+----------------+---|
            return shiftBits(clampBits(ratioT, 6), 0) |
                   shiftBits(clampBits(f1LargerThanF2, 1), 6) |
                   shiftBits(clampBits(f2LargerThanF3, 1), 7) |
                   shiftBits(clampBits(f3LargerThanF1, 1), 8) |
                   shiftBits(clampBits(m1LargerThanM2, 1), 9) |
                   shiftBits(clampBits(m2LargerThanM3, 1), 10) |
                   shiftBits(clampBits(m3LargerThanM1, 1), 11) |
                   shiftBits(clampBits(dt1t2LargerThant3t2, 1), 12) |
                   shiftBits(clampBits(df1f2LargerThanf3f2, 1), 13) |
                   shiftBits(clampBits(f1Range, 6), 14) |
                   shiftBits(clampBits(uint32_t(df2f1), 6), 20) |
                   shiftBits(clampBits(uint32_t(df3f2), 6), 26);
        }
    };

    std::vector<Fingerprint> packEventPoints() {
        std::vector<Fingerprint> results;

        const auto minFreqDistance = PANAKO_FP_MIN_FREQ_DIST;
        const auto maxFreqDistance = PANAKO_FP_MAX_FREQ_DIST;

        const auto minTimeDistance = PANAKO_FP_MIN_TIME_DIST;
        const auto maxTimeDistance = PANAKO_FP_MAX_TIME_DIST;

        for (auto i = 0; i < points.size(); ++i) {
            auto& p1 = points[i];
            for (auto j = i + 1; j < points.size(); ++j) {
                auto& p2 = points[j];

                auto fDiff = std::abs(int64_t(p1.frequency) - int64_t(p2.frequency));
                auto tDiff = int64_t(p2.time) - int64_t(p1.time);

                if (tDiff > maxTimeDistance) {
                    break;
                }
                if (tDiff < minTimeDistance) {
                    continue;
                }

                if (fDiff < minFreqDistance) {
                    continue;
                }
                if (fDiff > maxFreqDistance) {
                    continue;
                }

                for (auto k = j + 1; k < points.size(); ++k) {
                    auto& p3 = points[k];

                    fDiff = std::abs(int64_t(p2.frequency) - int64_t(p3.frequency));
                    tDiff = int64_t(p3.time) - int64_t(p2.time);

                    if (tDiff > maxTimeDistance) {
                        break;
                    }
                    if (tDiff < minTimeDistance) {
                        continue;
                    }

                    if (fDiff < minFreqDistance) {
                        continue;
                    }
                    if (fDiff > maxFreqDistance) {
                        continue;
                    }

                    results.emplace_back(p1, p2, p3);
                }
            }
        }

        return results;
    }

    std::vector<float> audioSamplesOut;
    size_t audioSamplesOutIndex = 0;
    const int sampleRate;
    const int channels;
    const int timeResolution;
    std::vector<uint8_t> byte_result;
    int magnitudesIndex = 0;
    int analysisFrameIndex = 0;
    std::vector<float> maxHorizontal;
    std::vector<EventPoint> points;
    std::vector<std::vector<float>> magnitudes;
    std::vector<std::vector<float>> maxMagnitudes;
    std::unordered_map<int,std::vector<float>*> previousMaxMagnitudes;
    std::unordered_map<int,std::vector<float>*> previousMagnitudes;
    LemireMaxClampedFilter* maxFilterVertical;
    SRC_STATE* src;
    uintptr_t gaborator;
    int64_t latency;
    int bandNumber;
    int sampleCount;
};



uintptr_t new_point_processor(int sampleRate, int channels, int time_resolution_in_ms){
    return reinterpret_cast<uintptr_t>(new EventPointProcessor(sampleRate, channels, time_resolution_in_ms));
}

void point_processor_process(uintptr_t ptr, float* data, size_t data_size){
    reinterpret_cast<EventPointProcessor*>(ptr)->Process(data, data_size);
}

const uint8_t* point_processor_finish(uintptr_t ptr, uint32_t use_compact_hash, uint32_t use_packed_print, size_t* result_size){
    auto& result = reinterpret_cast<EventPointProcessor*>(ptr)->Finish(use_compact_hash, use_packed_print);
    *result_size = result.size();
    return result.data();
}

void delete_point_processor(uintptr_t ptr){
    delete reinterpret_cast<EventPointProcessor*>(ptr);
}